@@ -167,7 +167,7 @@ zsh_gethostbyname2(char const *name, int af)
 
 /**/
 mod_export struct hostent *
-zsh_getipnodebyname(char const *name, int af, int flags, int *errorp)
+zsh_getipnodebyname(char const *name, int af, UNUSED(int flags), int *errorp)
 {
     static struct hostent ahe;
     static char nbuf[16];
@@ -195,7 +195,7 @@ zsh_getipnodebyname(char const *name, int af, int flags, int *errorp)
 
 /**/
 mod_export void
-freehostent(struct hostent *ptr)
+freehostent(UNUSED(struct hostent *ptr))
 {
 }
 
@@ -216,7 +216,7 @@ zts_alloc(int ztflags)
 {
     Tcp_session sess;
 
-    sess = (Tcp_session)zcalloc(sizeof(struct tcp_session));
+    sess = (Tcp_session)zshcalloc(sizeof(struct tcp_session));
     if (!sess) return NULL;
     sess->fd=-1;
     sess->flags=ztflags;
@@ -300,7 +300,7 @@ tcp_close(Tcp_session sess)
 	{
 	    err = close(sess->fd);
 	    if (err)
-		zwarn("connection close failed: %e", NULL, errno);
+		zwarn("connection close failed: %e", errno);
 	}
 	zts_delete(sess);
 	return 0;
@@ -337,10 +337,10 @@ tcp_connect(Tcp_session sess, char *addrp, struct hostent *zhost, int d_port)
 }
 
 static int
-bin_ztcp(char *nam, char **args, Options ops, int func)
+bin_ztcp(char *nam, char **args, Options ops, UNUSED(int func))
 {
     int herrno, err=1, destport, force=0, verbose=0, test=0, targetfd=0;
-    SOCKLEN_T  len;
+    ZSOCKLEN_T  len;
     char **addrp, *desthost, *localname, *remotename;
     struct hostent *zthost = NULL, *ztpeer = NULL;
     struct servent *srv;
@@ -358,8 +358,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
     if (OPT_ISSET(ops,'d')) {
 	targetfd = atoi(OPT_ARG(ops,'d'));
 	if (!targetfd) {
-	    zwarnnam(nam, "%s is an invalid argument to -d",
-		     OPT_ARG(ops,'d'), 0);
+	    zwarnnam(nam, "%s is an invalid argument to -d", OPT_ARG(ops,'d'));
 	    return 1;
 	}
     }
@@ -373,7 +372,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	    targetfd = atoi(args[0]);
 	    sess = zts_byfd(targetfd);
 	    if(!targetfd) {
-		zwarnnam(nam, "%s is an invalid argument to -c", args[0], 0);
+		zwarnnam(nam, "%s is an invalid argument to -c", args[0]);
 		return 1;
 	    }
 
@@ -381,7 +380,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	    {
 		if ((sess->flags & ZTCP_ZFTP) && !force)
 		{
-		    zwarnnam(nam, "use -f to force closure of a zftp control connection", NULL, 0);
+		    zwarnnam(nam, "use -f to force closure of a zftp control connection");
 		    return 1;
 		}
 		tcp_close(sess);
@@ -389,7 +388,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	    }
 	    else
 	    {
-		zwarnnam(nam, "fd %s not found in tcp table", args[0], 0);
+		zwarnnam(nam, "fd %s not found in tcp table", args[0]);
 		return 1;
 	    }
 	}
@@ -398,7 +397,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	int lport = 0;
 
 	if (!args[0]) {
-	    zwarnnam(nam, "-l requires an argument", NULL, 0);
+	    zwarnnam(nam, "-l requires an argument");
 	    return 1;
 	}
 
@@ -407,13 +406,13 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	    lport = srv->s_port;
 	else
 	    lport = htons(atoi(args[0]));
-	if (!lport) { zwarnnam(nam, "bad service name or port number", NULL, 0);
+	if (!lport) { zwarnnam(nam, "bad service name or port number");
 	return 1;
 	}
 	sess = tcp_socket(PF_INET, SOCK_STREAM, 0, ZTCP_LISTEN);
 
 	if (!sess) {
-	    zwarnnam(nam, "unable to allocate a TCP session slot", NULL, 0);
+	    zwarnnam(nam, "unable to allocate a TCP session slot");
 	    return 1;
 	}
 #ifdef SO_OOBINLINE
@@ -422,7 +421,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 #endif
 	if (!zsh_inet_aton("0.0.0.0", &(sess->sock.in.sin_addr)))
 	{
-	    zwarnnam(nam, "bad address: %s", "0.0.0.0", 0);
+	    zwarnnam(nam, "bad address: %s", "0.0.0.0");
 	    return 1;
 	}
 
@@ -433,7 +432,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	if (bind(sess->fd, (struct sockaddr *)&sess->sock.in, sizeof(struct sockaddr_in)))
 	{
 	    char buf[DIGBUFSIZE];
-	    convbase(buf, (zlong)lport, 10);
+	    convbase(buf, (zlong)ntohs(lport), 10);
 	    zwarnnam(nam, "could not bind to port %s: %e", buf, errno);
 	    tcp_close(sess);
 	    return 1;
@@ -441,7 +440,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 
 	if (listen(sess->fd, 1))
 	{
-	    zwarnnam(nam, "could not listen on socket: %e", NULL, errno);
+	    zwarnnam(nam, "could not listen on socket: %e", errno);
 	    tcp_close(sess);
 	    return 1;
 	}
@@ -468,26 +467,26 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	int lfd, rfd;
 
 	if (!args[0]) {
-	    zwarnnam(nam, "-a requires an argument", NULL, 0);
+	    zwarnnam(nam, "-a requires an argument");
 	    return 1;
 	}
 
 	lfd = atoi(args[0]);
 
 	if (!lfd) {
-	    zwarnnam(nam, "invalid numerical argument", NULL, 0);
+	    zwarnnam(nam, "invalid numerical argument");
 	    return 1;
 	}
 
 	sess = zts_byfd(lfd);
 	if (!sess) {
-	    zwarnnam(nam, "fd %s is not registered as a tcp connection", args[0], 0);
+	    zwarnnam(nam, "fd %s is not registered as a tcp connection", args[0]);
 	    return 1;
 	}
 
 	if (!(sess->flags & ZTCP_LISTEN))
 	{
-	    zwarnnam(nam, "tcp connection not a listener", NULL, 0);
+	    zwarnnam(nam, "tcp connection not a listener");
 	    return 1;
 	}
 
@@ -502,7 +501,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	    if ((ret = poll(&pfd, 1, 0)) == 0) return 1;
 	    else if (ret == -1)
 	    {
-		zwarnnam(nam, "poll error: %e", NULL, errno);
+		zwarnnam(nam, "poll error: %e", errno);
 		return 1;
 	    }
 # else
@@ -515,17 +514,17 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	    tv.tv_sec = 0;
 	    tv.tv_usec = 0;
 	    
-	    if (ret = select(lfd+1, &rfds, NULL, NULL, &tv)) return 1;
+	    if ((ret = select(lfd+1, &rfds, NULL, NULL, &tv))) return 1;
 	    else if (ret == -1)
 	    {
-		zwarnnam(nam, "select error: %e", NULL, errno);
+		zwarnnam(nam, "select error: %e", errno);
 		return 1;
 	    }
 	    
 # endif
 	    
 #else
-	    zwarnnam(nam, "not currently supported", NULL, 0);
+	    zwarnnam(nam, "not currently supported");
 	    return 1;
 #endif
 	}
@@ -534,7 +533,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	len = sizeof(sess->peer.in);
 	if ((rfd = accept(lfd, (struct sockaddr *)&sess->peer.in, &len)) == -1)
 	{
-	    zwarnnam(nam, "could not accept connection: %e", NULL, errno);
+	    zwarnnam(nam, "could not accept connection: %e", errno);
 	    tcp_close(sess);
 	    return 1;
 	}
@@ -571,7 +570,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 		    if (ztpeer)
 			remotename = ztpeer->h_name;
 		    else
-			remotename = ztrdup(inet_ntoa(sess->peer.in.sin_addr));
+			remotename = inet_ntoa(sess->peer.in.sin_addr);
 		    if (OPT_ISSET(ops,'L')) {
 			int schar;
 			if (sess->flags & ZTCP_ZFTP)
@@ -615,14 +614,14 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	
 	zthost = zsh_getipnodebyname(desthost, AF_INET, 0, &herrno);
 	if (!zthost || errflag) {
-	    zwarnnam(nam, "host resolution failure: %s", desthost, 0);
+	    zwarnnam(nam, "host resolution failure: %s", desthost);
 	    return 1;
 	}
 	
 	sess = tcp_socket(PF_INET, SOCK_STREAM, 0, 0);
 
 	if (!sess) {
-	    zwarnnam(nam, "unable to allocate a TCP session slot", NULL, 0);
+	    zwarnnam(nam, "unable to allocate a TCP session slot");
 	    return 1;
 	}
 
@@ -632,7 +631,7 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 #endif
 
 	if (sess->fd < 0) {
-	    zwarnnam(nam, "socket creation failed: %e", NULL, errno);
+	    zwarnnam(nam, "socket creation failed: %e", errno);
 	    zsfree(desthost);
 	    zts_delete(sess);
 	    return 1;
@@ -640,14 +639,14 @@ bin_ztcp(char *nam, char **args, Options ops, int func)
 	
 	for (addrp = zthost->h_addr_list; err && *addrp; addrp++) {
 	    if (zthost->h_length != 4)
-		zwarnnam(nam, "address length mismatch", NULL, 0);
+		zwarnnam(nam, "address length mismatch");
 	    do {
 		err = tcp_connect(sess, *addrp, zthost, destport);
 	    } while (err && errno == EINTR && !errflag);
 	}
 	
 	if (err) {
-	    zwarnnam(nam, "connection failed: %e", NULL, errno);
+	    zwarnnam(nam, "connection failed: %e", errno);
 	    tcp_close(sess);
 	    zsfree(desthost);
 	    return 1;
@@ -676,21 +675,44 @@ static struct builtin bintab[] = {
     BUILTIN("ztcp", 0, bin_ztcp, 0, 3, 0, "acd:flLtv", NULL),
 };
 
+static struct features module_features = {
+    bintab, sizeof(bintab)/sizeof(*bintab),
+    NULL, 0,
+    NULL, 0,
+    NULL, 0,
+    0
+};
+
 /* The load/unload routines required by the zsh library interface */
 
 /**/
 int
-setup_(Module m)
+setup_(UNUSED(Module m))
 {
     return 0;
 }
 
+/**/
+int
+features_(Module m, char ***features)
+{
+    *features = featuresarray(m, &module_features);
+    return 0;
+}
+
+/**/
+int
+enables_(Module m, int **enables)
+{
+    return handlefeatures(m, &module_features, enables);
+}
+
 /**/
 int
 boot_(Module m)
 {
     ztcp_sessions = znewlinklist();
-    return !addbuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab));
+    return 0;
 }
 
 
@@ -699,14 +721,13 @@ int
 cleanup_(Module m)
 {
     tcp_cleanup();
-    deletebuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab));
     freelinklist(ztcp_sessions, (FreeFunc) ztcp_free_session);
-    return 0;
+    return setfeatureenables(m, &module_features, NULL);
 }
 
 /**/
 int
-finish_(Module m)
+finish_(UNUSED(Module m))
 {
     return 0;
 }
